# -*- coding: utf-8 -*-
'''\
Unit tests for bpt.box
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os
from tempfile import mktemp
from shutil import rmtree

from nose.tools import assert_raises, with_setup

from bpt import UserError
from bpt.box import Box
from bpt.package import Package

def setup():
    global box1_path
    box1_path = mktemp()

def rmbox():
    try:
        rmtree(box1_path)
    except OSError:
        pass

@with_setup(teardown=rmbox)
def test_create():
    assert_raises(UserError, Box, box1_path) # No box, should raise
    box = Box.create(box1_path)
    assert box.path == box1_path
    assert_raises(UserError, Box.create, box1_path)

@with_setup(teardown=rmbox)
def test_packages():
    box = Box.create(box1_path)
    foo = box.create_package('foo-0.1', app_name='foo', app_version='0.1', enabled=True)
    bar = box.create_package('bar-0.2', app_name='bar', app_version='0.2', enabled=False)
    
    assert len(list(box.packages())) == 2
    l = list(box.packages(only_enabled=True))
    assert len(l) == 1
    assert l[0] is foo
    
    box.disable_package(foo)
    assert len(list(box.packages(only_enabled=True))) == 0
    
    box.enable_package(foo)
    box.enable_package(bar)
    assert len(list(box.packages(only_enabled=True))) == 2
    assert len(list(box.packages(only_enabled=True, matching=['foo.*']))) == 1

    box.disable_package(foo)
    assert len(list(box.packages(only_enabled=True, matching=['foo.*']))) == 0
    assert len(list(box.packages(matching=['foo.*']))) == 1
    
    box.disable_package(bar, remove=True)
    assert len(list(box.packages())) == 1
    
    
    
    
    
    
    
    
    



# -*- coding: utf-8 -*-
'''\
Command dispatching framework for BPT.
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import re

from bpt.ui.config import BPTOptionParser

# XXX(ot): is there something standard for this?
class classproperty(object):
    def __init__(self, fget):
        self._fget = fget

    def __get__(self, obj, objtype):
        return self._fget(objtype)

# This code is inspired by jhbuild's commands framework
class Command(object):
    '''Base class for Command objects.'''
    
    def __init__(self, options=None):
        if options is None:
            options = []
        self.options = options

    # Override the following 
    # __doc__ = ...
    usage_args = '[ options ... ]'

    @classproperty
    def name(cls):
        return cls.__name__
    
    @classproperty
    def description(cls):
        return re.sub('\s+', ' ', cls.__doc__)

    def execute(self, config, cmd_args):
        options, pos_args = self._parse_args(cmd_args)
        try:
            return self._run(config, options, pos_args)
        except InvalidCommandLine:
            self.parser.print_help()
            return 1

    def _parse_args(self, args):
        self.parser = BPTOptionParser(
            usage='%%prog %s %s' % (self.name, self.usage_args),
            description=self.description)
        self.parser.add_options(self.options)
        return self.parser.parse_args(args)

    def _require_args(self, cmd_args, at_least=0, at_most=None):
        l = len(cmd_args)
        if l < at_least:
            raise InvalidCommandLine
        if at_most is not None and l > at_most:
            raise InvalidCommandLine
        

    def _run(self, config, options, pos_args):
        raise NotImplementedError

class InvalidCommandLine(Exception): pass

class CommandNotFound(Exception): pass

def get_commands():
    return Command.__subclasses__() # pylint: disable-msg=E1101

def dispatch(command, config, cmd_args):
    '''Run a command. Raises CommandNotFound if command does not exist'''
    
    for cmd_class in get_commands():
        if cmd_class.__name__ == command:
            break
    else:
        raise CommandNotFound
    
    cmd = cmd_class()
    ret = cmd.execute(config, cmd_args)
    if ret is None: ret = 0
    return ret




# -*- coding: utf-8 -*-
'''\
Package class
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os

from bpt import UserError
from bpt.util import store_info, load_info

class Package(object):
    '''A package installed in a box, i.e. a subdirectory of
    <box>/pkgs. The attributes are kept in sync with the stored
    pkg_info'''

    _PackagePool = dict()

    def __new__(cls, pkgdir):
        # Ensure that there is only one instance of Package for every
        # directory, so that no other instance can change pkg_info
        # contents, invalidating our _dict. 

        # XXX(ot): This is obviously not thread safe
        # XXX(ot): This keeps all the packages in memory. Use a
        # weakrefdict instead?

        return Package._PackagePool.setdefault(pkgdir, object.__new__(cls))
        

    def __init__(self, pkgdir):
        # XXX(ot): better exceptions? metadata version handling?

        self._path = pkgdir
        self._name = os.path.basename(pkgdir)
        self._dict = load_info(_pkg_info_file(pkgdir))

        # Sanity check
        for k in ['app_name', 'app_version', 'enabled']:
            if k not in self._dict:
                raise UserError('Invalid package')

    @classmethod
    def create(cls, pkgdir, **kwArgs):
        try:
            os.makedirs(os.path.join(pkgdir, 'bpt_meta'))
        except OSError, exc:
            if exc.errno != 17: # directory exists
                raise

        store_info(_pkg_info_file(pkgdir), kwArgs)
        return cls(pkgdir)
        

    def __getattr__(self, attr):
        try:
            return self._dict[attr]
        except KeyError, exc:
            raise AttributeError(str(exc))

    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            return object.__setattr__(self, attr, value)

        self._dict[attr] = value
        store_info(_pkg_info_file(self._path), self._dict)

    @property
    def path(self):
        return self._path

    @property
    def name(self):
        return self._name

    def __str__(self):
        return self.name

def _pkg_info_file(pkgdir):
    return os.path.join(pkgdir, 'bpt_meta', 'pkg_info')



# -*- coding: utf-8 -*-
'''\
Unit tests for bpt.build
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os
from tempfile import mktemp
from shutil import rmtree

from nose.tools import assert_raises, with_setup

from bpt import UserError
from bpt.box import Box
from bpt.build import SourceDir

def setup():
    global box1_path
    global box2_path
    box1_path = mktemp()
    box2_path = mktemp()

    global sourcedirs_path
    sourcedirs_path = os.path.join(os.path.dirname(__file__), 'fake_sourcedirs')

def rmbox():
    try:
        rmtree(box1_path)
        rmtree(box2_path)
    except OSError:
        pass
def test_sourcedir():
    assert_raises(UserError, SourceDir, box1_path) # Not a sourcedir, should raise

@with_setup(teardown=rmbox)
def test_build():
    box = Box.create(box1_path)
    box2 = Box.create(box2_path)

    foo_path = os.path.join(sourcedirs_path, 'foo')
    foo_source = SourceDir(foo_path)
    assert foo_source.path == foo_path
    bar_path = os.path.join(sourcedirs_path, 'bar')
    bar_source = SourceDir(bar_path)

    foo = foo_source.build(box)
    bar = bar_source.build(box)
    foo_new = foo_source.build(box, name_suffix='-new')

    assert foo.enabled == False
    assert foo_new.enabled == True

    assert os.path.samefile(os.path.join(box.virtual_path, 'lib', 'd', 'foo'),
                            os.path.join(foo_new.path, 'lib', 'd', 'foo'))
    
    assert foo_source.unittest() == True
    assert bar_source.unittest() == False
    
    foo_source.build(box2) # Should clean before

    assert len(list(box.packages(only_enabled=True))) == 2

    box.disable_package(foo_new)
    assert not os.path.exists(os.path.join(box.virtual_path, 'lib', 'd', 'foo'))

    foo_source.clean()
    bar_source.clean()



# -*- coding: utf-8 -*-
'''\
autoremove command
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import re
from optparse import make_option

from bpt import UserError, log
from bpt.ui.command import Command
from bpt.box import require_box

class autoremove(Command):
    '''Remove disabled packages not matching the given regexps'''

    usage_args = ''

    def __init__(self):
        options = [make_option('-e', '--exclude', action='append',
                               dest='exclude',
                               default=[],
                               help='Do not remove packages matching REGEX. Can be specified multiple times.',
                               metavar='REGEX')
                   ]
        Command.__init__(self, options)
        

    def _run(self, config, cmd_options, cmd_args):
        require_box(config)
        box = config.box

        regexps = [re.compile(pattern + '$', re.IGNORECASE) for pattern in cmd_options.exclude]

        to_remove = []
        for pkg in box.packages():
            if pkg.enabled: 
                continue
            for regexp in regexps:
                if regexp.match(pkg.name):
                    break
            else: # no exclude regexp matched
                to_remove.append(pkg)

        if not to_remove:
            log.info('No packages to remove')
            return 0

        print 'The following packages will be removed:'
        for pkg in to_remove:
            print '\t%s' % pkg.name
        answer = raw_input('Are you sure [y/N]? ')
        if answer.lower() != 'y':
            return 0
        for pkg in to_remove:
            box.disable_package(pkg, remove=True)
                



# -*- coding: utf-8 -*-
'''\
Command line frontend for BPT
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import logging

import bpt
from bpt import log
from bpt import ui
from bpt.ui.config import Config, BPTOptionParser, ParserExit
from bpt.box import Box, get_current_box

def help_commands(option, opt_str, value, parser):
    commands = [(x.name, x.description) for x in ui.command.get_commands()]
    commands.sort()
    # XXX(ot): use more formatting magic for long descriptions?
    for name, description in commands:
        print '  %-15s %s' % (name, description)
    print
    print 'For more informations about <command> run "%s <command> --help"' % parser.get_prog_name()
    parser.exit()

def setup_optparse():
    parser = BPTOptionParser(
        usage='%prog command [ options ... ]',
        description='Tool for creating/managing package boxes',
        version='%%prog %s' % bpt.__version__)
    parser.disable_interspersed_args()

    parser.add_option('--help-commands', action='callback',
                      callback=help_commands,
                      help='informations about available commands')
    parser.add_option('-b', '--box', action='store',
                      type='string', dest='box_path',
                      help='box path. If not specified, use the current one.')

    return parser
    
def main(argv, do_log=True):
    # set up default logging 
    # TODO(giuott): Add a verbosity option
    if do_log:
        logging.basicConfig()
        log.setLevel(logging.INFO)
    
    try:
        parser = setup_optparse()
        options, args = parser.parse_args(argv[1:])
    except ParserExit:
        return 0
    
    if not args:
        parser.print_help()
        return 255

    command = args[0]
    cmd_args = args[1:]
    config = Config()

    try:
        if options.box_path is not None:
            config.box = Box(options.box_path)
        else:
            config.box = get_current_box()
            if config.box is not None:
                log.info('Using current box "%s"', config.box.name)

        return ui.command.dispatch(command, config, cmd_args)
    except ui.command.CommandNotFound:
        log.error("Command %s not found", command)
        parser.print_help()
        return 255
    except bpt.UserError, exc:
        log.error('Aborting: %s', str(exc))
        return 1
    except ParserExit:
        return 0
    



# -*- coding: utf-8 -*-
'''\
Utility script to automate installation of packages with common build patterns
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os
import sys
import re
from tempfile import mkdtemp
from subprocess import check_call
from shutil import rmtree

from bpt import log, UserError

SETUPTOOLS_WORKAROUND = True 

class UnsupportedPath(Exception): 
    '''Unable to guess file extension, package metadata or how to
    build the package'''

def guess_unpack_cmd(filename):
    '''
    >>> guess_unpack_cmd('/tmp/python-2.5.tar.gz')
    ('python-2.5', ['tar', 'zsxf', '/tmp/python-2.5.tar.gz'])
    
    >>> guess_unpack_cmd('/tmp/unsupported.7z')
    Traceback (most recent call last):
    ...
    UnsupportedPath: Unsupported file extension
    '''

    cmd = []
    root = os.path.basename(filename)
    root, ext = os.path.splitext(root)
    if ext == '.gz':
        root, ext = os.path.splitext(root)
        if ext != '.tar':
            raise UnsupportedPath('Unsupported extension')
        cmd = ['tar', 'zsxf']
    elif ext == '.bz2':
        root, ext = os.path.splitext(root)
        if ext != '.tar':
            raise UnsupportedPath('Unsupported extension')
        cmd = ['tar', 'jsxf']
    elif ext == '.tar':
        cmd = ['tar', 'sxf']
    elif ext == '.tgz':
        cmd = ['tar', 'zsxf']
    elif ext == '.zip':
        cmd = ['unzip']
    else:
        raise UnsupportedPath('Unsupported file extension')

    cmd.append(os.path.abspath(filename))
    
    return root, cmd

FILENAME_RE = re.compile(r'^(.*)-(\d.*)$')
def guess_name_version(basename):
    '''
    >>> guess_name_version('python-2.5')
    ('python', '2.5')

    >>> guess_name_version('abc.25')
    Traceback (most recent call last):
    ...
    UnsupportedPath: Unable to guess name and version from "abc.25"
    '''

    match = FILENAME_RE.match(basename)
    if not match:
        raise UnsupportedPath('Unable to guess name and version from "%s"' % basename)
    name, version = match.groups()
    return name, version

def autobuild_dir(box, source_dir, basename, configure_options, stdout=None):

    name, version = guess_name_version(basename)
    log.info('Guessed application name "%s", version "%s"', name, version)

    # XXX(ot): disable package if something goes wrong?
    pkg = box.create_package(basename, 
                             app_name=name, 
                             app_version=version,
                             enabled=False)

    if os.path.exists(os.path.join(source_dir, 'configure')):
        cmd_list = [
            "./configure --prefix '%s' %s" % (pkg.path, configure_options),
            "make install"
            ]
    elif os.path.exists(os.path.join(source_dir, 'setup.py')):
        envvars = ''
        if SETUPTOOLS_WORKAROUND:
            # Workaround two bugs of setuptools 
            # (fixed in http://bugs.python.org/setuptools/issue54 )
            # - Put the site-packages path in python's path
            # - Create the site-packages directory

            # XXX(ot): find a better solution
            # XXX(ot): this assumes that python inside the box is the same as the current
            python_version = '.'.join([str(x) for x in sys.version_info[:2]])
            site_path = '%s/lib/python%s/site-packages/' % (pkg.path, python_version)
            envvars = 'PYTHONPATH="%s${PYTHONPATH:+:$PYTHONPATH}"' % site_path
            try: 
                os.makedirs(site_path)
            except OSError: 
                pass
        cmd_list = [
            "%s python setup.py install --prefix '%s' %s" % 
            (envvars, pkg.path, configure_options),
            ]
    else:
        raise UnsupportedPath('Do not know how to build source')

    log.info('Building and installing as package %s', pkg.name)
    for cmd in cmd_list:
        check_call(['bash', '-e', box.env_script, cmd], cwd=source_dir, stdout=stdout)

    box.enable_package(pkg)
    

def autobuild(box, filename, configure_options='', keep_temp=False, stdout=None):
    if not os.path.exists(filename):
        raise UserError('File %s does not exist' % filename)
    if os.path.isdir(filename):
        source_dir = os.path.abspath(filename)
        autobuild_dir(box, source_dir, os.path.basename(source_dir), configure_options)
    else:
        build_dir = mkdtemp()
        try:
            basename, unpack_cmd = guess_unpack_cmd(filename)

            log.info('Unpacking the file...')
            check_call(unpack_cmd, cwd=build_dir)

            unpacked_files = [os.path.join(build_dir, f) 
                              for f in os.listdir(build_dir)]
            unpacked_dirs = [d for d in unpacked_files if os.path.isdir(d)]
            if len(unpacked_dirs) != 1:
                raise UnsupportedPath('Could not find source directory')
            source_dir, = unpacked_dirs
            autobuild_dir(box, source_dir, basename, configure_options, stdout)
        finally:
            if keep_temp:
                log.info('Not deleting temporary files in directory %s', build_dir)
            else:
                rmtree(build_dir)
    

        



# -*- coding: utf-8 -*-
'''\
Utility functions to link (and unlink) recursively directories
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os

from bpt import log
from bpt.util import path_diff

def linkdir(src_path, dst_path):
    rel_src_path, _ = path_diff(dst_path, src_path)
    
    for root, dirs, files in os.walk(src_path):
        rel_path, rel_parents = path_diff(src_path, root)
        
        for d in dirs:
            dpath = os.path.join(dst_path, rel_path, d)
            if not os.path.exists(dpath):
                os.makedirs(dpath)
            else:
                assert os.path.isdir(dpath), '%d is a directory in package and a non-directory in destination' % dpath
        
        for f in files:
            spath = os.path.join(rel_parents, rel_src_path, rel_path, f)
            dpath = os.path.join(dst_path, rel_path, f)
            if os.path.lexists(dpath):
                if os.path.islink(dpath):
                    os.unlink(dpath)
                else:
                    log.warn('Existing file (not a symlink): %s' % dpath)
                    continue
            
            os.symlink(spath, dpath)

def unlinkdir(src_path, dst_path):
    for root, dirs, files in os.walk(src_path, topdown=False):
        rel_path, rel_parents = path_diff(src_path, root)
        
        for d in dirs:
            dpath = os.path.join(dst_path, rel_path, d)
            try:
                os.rmdir(dpath)
            except OSError:
                # Directory not empty, leaving it
                pass
        
        for f in files:
            dpath = os.path.join(dst_path, rel_path, f)
            if not os.path.exists(dpath): 
                continue

            if os.path.islink(dpath):
                if os.path.samefile(dpath, os.path.join(root, f)):
                    os.unlink(dpath)
            else:
                log.warn('%s not a symlink, not removing', dpath)





# -*- coding: utf-8 -*-
'''\
Box class implementation
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os
import re
import shutil
from uuid import uuid1
from tempfile import mktemp # for doctests, pylint: disable-msg=W0611

import bpt
from bpt import log, UserError
from bpt.util import store_info, load_info
from bpt.linkdir import linkdir, unlinkdir
from bpt.package import Package

# Directories created inside the box
STANDARD_DIRS = ['pkgs', 'bpt_meta', 'bin', 'lib', 'man', 'share', 'include']

# Directories whose contents are linked from every package to the box
DYN_DIRS = ['bin', 'lib', 'man', 'share', 'include']

STANDARD_PATH_VARS = [('PATH', 'bin'), 
                      ('LIBRARY_PATH', 'lib'), 
                      ('LD_LIBRARY_PATH', 'lib'),
                      ('DYLD_LIBRARY_PATH', 'lib'), 
                      ('CPATH', 'include'), 
                      ('MANPATH', 'man'), 
                      ('PKG_CONFIG_PATH', os.path.join('lib', 'pkgconfig'))]

ENV_SCRIPT_TMPL = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'env_script.tmpl'))

class Box(object):
    '''\
    Represents a package box in the filesystem. 

    The constructor expects an existing box. A new one can be created
    with the class method create()
    '''

    def __init__(self, box_path):
        self._path = os.path.abspath(box_path)

        try:
            box_info = load_info(os.path.join(self.path, 
                                              'bpt_meta', 
                                              'box_info'))
        except (OSError, IOError):
            raise UserError('Invalid box: impossible to read box_info')

        try:
            self._id = box_info['id']
            self._platform = box_info['platform']
        except KeyError, exc:
            raise UserError('Invalid box_info: missing "%s"', str(exc))
        
        # Ensure that the virtual path symlink is existing and points
        # to the correct location
        try:
            os.symlink(self.path, self.virtual_path)
        except OSError, exc:
            if exc.errno == 17 and \
                    not os.path.samefile(self.path, self.virtual_path):
                raise UserError('Virtual path symlink %s exists but ' 
                                'does not point to this box. Please remove '
                                'it manually', self.virtual_path)

    def __repr__(self):
        return "Box('%s')" % self.path
    
    @property
    def box_id(self):
        '''Unique id of the box, used to build the virtual path'''
        return self._id
    
    @property
    def path(self):
        '''Actual path of the box'''
        return self._path

    @property
    def virtual_path(self):
        '''Virtual (or relocatable) path of the box. It is a symlink
        that points to the box, but does not depend on the actual path'''
        return os.path.join('/tmp', 'box_' + self.box_id)

    @property
    def name(self):
        '''The name of the box, defined as the basename of the directory'''
        return os.path.basename(os.path.abspath(self.path))

    @property
    def env_script(self):
        '''Path of the env script'''
        return os.path.join(self.path, 'env')

    def __eq__(self, other):
        ''' Compare equality by box_id

        >>> box1 = Box.create(mktemp())
        >>> box2 = Box.create(mktemp())
        >>> box1 == box1
        True
        >>> box1 == box2
        False
        '''
        return isinstance(other, Box) and self.box_id == other.box_id

    @classmethod
    def create(cls, dest_path):
        '''Creates a directory dest_path and initialize a package box
        in it.  Returns the initialized box.

        >>> box_path = mktemp()
        >>> box = Box.create(box_path)
        >>> box.path == box_path
        True
        '''
        # Safety checks
        if os.path.exists(dest_path):
            raise UserError('Destination already exists')

        try:
            os.makedirs(dest_path)
            for directory in STANDARD_DIRS:
                os.makedirs(os.path.join(dest_path, directory))
        except OSError, exc:
            raise UserError('Impossible to create destination directories: "%s"' % 
                            str(exc))

        box_info = dict()
        box_info['id'] = str(uuid1())
        box_info['platform'] = _get_platform()
        box_info['bpt_version'] = bpt.__version__
        store_info(os.path.join(dest_path, 'bpt_meta', 'box_info'), box_info)

        box = cls(dest_path)
        box.sync()
        log.info('Created box with id %s in directory %s', box.box_id, box.path)
        return box
    
    def sync(self):
        '''Recreate all the symlinks and the env script, restoring the
        consistency of the box.'''

        # Clean all symlinks
        for d in DYN_DIRS:
            d_path = os.path.join(self.path, d)
            shutil.rmtree(d_path)
            os.makedirs(d_path)

        # Relink all packages    
        for package in self.packages(only_enabled=True):
            self._link_package(package)

        self._create_env_script()
        log.info('Synchronized box')

    def get_package(self, pkgname):
        pkgs_dir = os.path.join(self.virtual_path, 'pkgs')
        return Package(os.path.join(pkgs_dir, pkgname))

    def packages(self, only_enabled=False, matching=None):
        '''Iterates on the installed packages. If only_enabled is
        True, only enabled packages are returned. matching can be a
        list of regexps, if it is given a package is returned only if
        it matches at least one regexp
        '''
        if matching is not None:
            regexps = [re.compile(pattern + '$', re.IGNORECASE) for pattern in matching]
        for pkgname in os.listdir(os.path.join(self.virtual_path, 'pkgs')):
            try:
                pkg = self.get_package(pkgname)
            except UserError:
                log.warning('Invalid entry in pkgs: %s', pkgname)
                continue
            if not only_enabled or pkg.enabled:
                if matching is None:
                    yield pkg
                else:
                    for regexp in regexps:
                        if regexp.match(pkg.name):
                            yield pkg
                            break
                            
    def create_package(self, pkg_name, **pkg_info):
        pkg_prefix = os.path.join(self.virtual_path, 'pkgs', pkg_name)

        # This creates also the directory pkg_prefix
        pkg = Package.create(pkgdir=pkg_prefix,
                             **pkg_info)
        return pkg


    def enable_package(self, package):

        # Disable other versions of the same application
        for other in self.packages(only_enabled=True):
            if other is not package and other.app_name.lower() == package.app_name.lower():
                self.disable_package(other)

        log.info('Enabling package %s', package)
        package.enabled = True
        self._link_package(package)
        self._create_env_script()
        

    def disable_package(self, package, remove=False):
        if not remove:
            log.info('Disabling package %s', package)
        else:
            log.info('Removing package %s', package)

        package.enabled = False
        self._unlink_package(package)
        self._create_env_script()
        if remove:
            shutil.rmtree(package.path)

    def check_platform(self):
        '''Check that the current platform is the same as box's one'''
        return self._platform == _get_platform()

    def _link_package(self, package):
        log.info('Linking package %s' % package.name)

        for d in DYN_DIRS:
            src_path = os.path.abspath(os.path.join(package.path, d))
            dest_path = os.path.abspath(os.path.join(self.virtual_path, d))
            linkdir(src_path, dest_path)

    def _unlink_package(self, package):
        log.info('Unlinking package %s' % package.name)

        for d in DYN_DIRS:
            src_path = os.path.abspath(os.path.join(package.path, d))
            dest_path = os.path.abspath(os.path.join(self.virtual_path, d))
            unlinkdir(src_path, dest_path)

    def _create_env_script(self):
        virtual_path = self.virtual_path
        path_updates = '\n'.join('export %s="$VIRTUAL_PATH/%s${%s:+:$%s}"' % (v, d, v, v) 
                                 for v, d in STANDARD_PATH_VARS)

        # Collect package env scripts
        pkg_env_scripts = []
        for pkg in self.packages(only_enabled=True):
            local_env_script = os.path.join(pkg.path, 'bpt_meta', 'env_script')
            if os.path.exists(local_env_script):
                f = open(local_env_script)
                try:
                    pkg_env_scripts.append(f.read())
                finally:
                    f.close()

        pkg_env_scripts = '\n'.join(pkg_env_scripts)
        
        env_script = open(ENV_SCRIPT_TMPL).read() % locals() # XXX(ot) close file?
        
        env_script_file = open(self.env_script, 'w')

        try: 
            env_script_file.write(env_script)
        finally:
            env_script_file.close()
        os.chmod(self.env_script, 0755)

        log.info('Created env script')


def require_box(config):
    '''Raise an exception if config.box is None'''
    if config.box is None:
        raise UserError('No box given')

def get_current_box():
    ''' If we are in a box environment, return the current box.
    Otherwise None.'''
    box_path = os.environ.get('BPT_BOX_PATH', None)
    if box_path is not None:
        try:
            box = Box(box_path)
            return box
        except UserError, exc:
            log.warning('Not using current box %s because of error "%s"', box_path, str(exc))
    return None
    
def _get_platform():
    u = os.uname()
    return (u[0], u[4])




# -*- coding: utf-8 -*-
'''\
Build command
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

from optparse import make_option

from bpt.ui.command import Command
from bpt.box import require_box
from bpt.build import SourceDir

class build(Command):
    '''Build a set of sourcedirs into the box'''

    usage_args = '<source package> ...'

    def __init__(self):
        options = [make_option('-c', '--clean-before', action='store_true',
                               dest='clean_before',
                               help='Clean sourcedir before building it.'),
                   make_option('-s', '--suffix', action='store',
                               dest='suffix',
                               default='',
                               help='Append a suffix to the package name.'),
                   make_option('-t', '--test', action='store_true',
                               dest='test',
                               help='Run the tests after building the package.')
                   ]
        Command.__init__(self, options)
        

    def _run(self, config, cmd_options, cmd_args):
        self._require_args(cmd_args, 1)
        require_box(config)

        for sourcedir in cmd_args:
            sd = SourceDir(sourcedir)
            if cmd_options.clean_before:
                sd.clean()
            sd.build(config.box, cmd_options.suffix)
            if cmd_options.test:
                sd.unittest()

class clean(Command):
    '''Clean a set of sourcedirs'''

    usage_args = '<source package> ...'

    def __init__(self):
        options = [make_option('-d', '--deep', action='store_true',
                               dest='deep',
                               help='Deep clean: erase also downloaded files, etc...')
                   ]
        Command.__init__(self, options)

    def _run(self, config, cmd_options, cmd_args):
        self._require_args(cmd_args, 1)

        for sourcedir in cmd_args:
            sd = SourceDir(sourcedir)
            sd.clean(cmd_options.deep)

class unittest(Command):
    '''Run unit tests inside a set of sourcedirs.
    Should be invoked only after a build command.'''

    usage_args = '<source package> ...'

    def _run(self, config, cmd_options, cmd_args):
        self._require_args(cmd_args, 1)

        for sourcedir in cmd_args:
            sd = SourceDir(sourcedir)
            sd.unittest()



# -*- coding: utf-8 -*-
'''\
Functions to build sourcedirs into a package box
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os
from subprocess import call

from bpt import log, UserError
from bpt.util import getstatusoutput, store_info, load_info

BASE_SH_SCRIPT = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 
                 'bpt_base_script.sh')
    )

class SourceDir(object):
    def __init__(self, sourcedir):
        self._sourcedir = sourcedir
        self._rulesfile = os.path.join(sourcedir, 'bpt-rules')
        if not os.path.exists(self._rulesfile):
            raise UserError('Invalid sourcedir')

    @property
    def path(self): return self._sourcedir
    
    def get_var(self, var, can_be_empty=False):
        # XXX(ot) find a better way to do this
        sh_line = ('cd %s;' % self._sourcedir
                   + 'source bpt-rules;'
                   + 'echo ${%s};' % var
                   )
        exitstatus, outtext = getstatusoutput("bash -e -c '%s'" % sh_line)
        assert exitstatus == 0, (exitstatus, outtext)

        outtext = outtext.strip()
        if outtext == '' and not can_be_empty:
            raise Exception('No variable %s' % var)

        return outtext

    def build(self, box, name_suffix='', stdout=None):
        appname = self.get_var('BPT_APP_NAME')
        version = self.get_var('BPT_APP_VERSION')

        # Check last box the sourcedir was built in
        bpt_status_file = os.path.join(self.path, '.bpt_status')
        if os.path.exists(bpt_status_file):
            bpt_status = load_info(bpt_status_file)
            last_box_id = bpt_status.get('last_box_id', box.box_id)
            if last_box_id != box.box_id: 
                log.info('Intermediate files built for another package box. Cleaning first.')
                self.clean()

        # Write the current box_id
        store_info(bpt_status_file, dict(last_box_id=box.box_id))
        
        if not box.check_platform():
            raise UserError('Current platform is different from box\'s platform')

        pkg_name = '%s-%s%s' % (appname, version, name_suffix)
        pkg = box.create_package(pkg_name,
                                 app_name=appname,
                                 app_version=version,
                                 enabled=False)
        log.info('Building application %s, in sourcedir %s', appname, self._sourcedir)

        # Build
        sh_line = ('source "%s";' % BASE_SH_SCRIPT
                   + 'cd "%s";' % self._sourcedir
                   + 'export BPT_PKG_PREFIX="%s";' % pkg.path
                   + 'source bpt-rules;'
                   + 'bpt_download;'
                   + 'bpt_unpack;'
                   + 'bpt_build;'
                   )
        retcode = call(['bash', '-e', box.env_script, sh_line], stdout=stdout)
        if retcode != 0:
            raise UserError('FATAL: build script exited with status %s' % retcode)

        box.enable_package(pkg)
        return pkg
        
    def clean(self, deep=False):
        log.info('Cleaning sourcedir %s', self.path)
        sh_line = ('source "%s";' % BASE_SH_SCRIPT
                   + 'cd "%s";' % self._sourcedir
                   + 'rm -f .bpt_status;'
                   + 'source bpt-rules;'
                   + 'bpt_clean;'
                   )
        if deep: 
            sh_line += 'bpt_deepclean;'
        exitstatus, outtext = getstatusoutput("bash -e -c '%s'" % sh_line)
        assert exitstatus == 0, (exitstatus, outtext)

    def unittest(self):
        log.info('Testing sourcedir %s', self.path)
        sh_line = ('source "%s";' % BASE_SH_SCRIPT
                   + 'cd "%s";' % self._sourcedir
                   + 'source bpt-rules;'
                   + 'bpt_unittest;'
                   )
        exitstatus = call(['bash', '-e', '-c', sh_line])
        if exitstatus != 0:
            log.warning('unittest exited with exit code %s. Some tests may have failed', exitstatus)
            return False
        return True
        
    



# -*- coding: utf-8 -*-
'''\
Configuration classes for BPT
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

from optparse import OptionParser

class ParserExit(Exception):
    '''Exception used by the option parser to signal program exit'''

class BPTOptionParser(OptionParser):
    def exit(self):
        raise ParserExit

class Config(object):
    '''Just a simple container of configuration values'''



# -*- coding: utf-8 -*-
'''\
Basic BPT commands
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os
from optparse import make_option
from subprocess import call

from bpt.ui.command import Command
from bpt.box import Box, require_box, get_current_box

class create(Command):
    '''Create a new package box'''

    usage_args = '<box path>'

    def _run(self, config, cmd_options, cmd_args):
        self._require_args(cmd_args, 1, 1)

        box_path = cmd_args[0]
        Box.create(box_path)

class sync(Command):
    '''Synchronize a package box'''

    usage_args = ''

    def _run(self, config, cmd_options, cmd_args):
        self._require_args(cmd_args, 0, 0)

        require_box(config)
        config.box.sync()

class shell(Command):
    '''Run a shell in the box\'s environment'''

    usage_args = ''

    def _run(self, config, cmd_options, cmd_args):
        self._require_args(cmd_args, 0, 0)
        require_box(config)

        box_name = '(%s)' % config.box.name
        shell_prompt = r'\[\033[01;31m\]%(box_name)s\[\033[01;32m\] \u@\h\[\033[01;34m\] \w \$\[\033[00m\] ' % locals()

        cur_box = get_current_box()
        if cur_box and cur_box == config.box:
            # Already inside the box env, just execute a shell with the new prompt
            new_env = dict(os.environ)
            new_env['PS1'] = shell_prompt
            call(['bash', '-norc', '-noprofile'], env=new_env)
        else:
            # Have to set the correct environment, so execute a shell using the env script
            # XXX(ot): does this spawn two bashs?
            call(['bash', config.box.env_script, 'PS1="%s"' % shell_prompt, 'bash', '-norc', '-noprofile'])

class status(Command):
    '''Show the installed packages. Packages can be narrowed down using a regex'''

    usage_args = '[pattern1 pattern2 ...]'

    def _run(self, config, cmd_options, cmd_args):
        require_box(config)
        
        fmt = '%-30s| %-20s| %-10s| %-10s|'

        print
        print fmt % ('PACKAGE', 'NAME', 'VERSION', 'STATUS')
        print 

        if len(cmd_args) == 0:
            patterns = None
        else:
            patterns = cmd_args 

        for package in config.box.packages(matching=patterns):
            if package.enabled:
                pkg_status = 'enabled'
            else:
                pkg_status = 'disabled'
            print fmt % (package.name,
                         package.app_name,
                         package.app_version,
                         pkg_status)
        print

class disable(Command):
    '''Disable the packages that match the patterns'''

    usage_args = '[pattern1 pattern2 ...]'

    def __init__(self):
        options = [make_option('-r', '--remove', action='store_true',
                               dest='remove',
                               help='Remove the packages after disabling them.')
                   ]
        Command.__init__(self, options)
        

    def _run(self, config, cmd_options, cmd_args):
        self._require_args(cmd_args, 1)
        require_box(config)

        patterns = cmd_args

        for package in config.box.packages(matching=patterns):
            config.box.disable_package(package, remove=cmd_options.remove)

class enable(Command):
    '''Enable the packages that match the patterns'''

    usage_args = '[pattern1 pattern2 ...]'

    def _run(self, config, cmd_options, cmd_args):
        self._require_args(cmd_args, 1)
        require_box(config)

        patterns = cmd_args

        for package in config.box.packages(matching=patterns):
            config.box.enable_package(package)



# -*- coding: utf-8 -*-
'''\
Common data for the BPT project
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

__version__ = '0.5.1'
__author__ = 'Giuseppe Ottaviano <giuott@gmail.com>'

import logging
log = logging.getLogger('BPT') # pylint: disable-msg=C0103

class UserError(Exception):
    '''Error to be reported to the user'''



# -*- coding: utf-8 -*-
'''\
Utility functions
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os
from pprint import pprint, isreadable
from tempfile import mktemp # for doctests, pylint: disable-msg=W0611

# Simple functions to store small dicts in human readable format
# WARNING: the stored data is evaluated, so it must be trusted
# XXX(ot): use another representation format?

def store_info(filename, data):
    r'''Pretty print roundtrippable data into filename
    
    >>> f = mktemp()
    >>> store_info(f, dict(a='hello', b=[1, 2, 3]))
    >>> open(f).read()
    "{'a': 'hello', 'b': [1, 2, 3]}\n"
    '''
    if not isreadable(data):
        raise Exception('Recursive or nonserializable data')
    
    f = open(filename, 'w')
    try:
        pprint(data, f)
    finally:
        f.close()

def load_info(filename):
    r'''Read data writen with store_info

    >>> f = mktemp()
    >>> data = dict(a='hello', b=[1, 2, 3])
    >>> store_info(f, data)
    >>> read_data = load_info(f)
    >>> read_data == data
    True
    '''
    f = open(filename)
    try:
        return eval(f.read())
    finally:
        f.close()
        
# Other util functions

def path_diff(start, path):
    '''Returns the relative path of a path from directory start, and
    the relative path to go back. Similar to Python 2.6 relpath, works
    only on POSIX.

    XXX(ot): Not sure this is robust or portable

    >>> path_diff('/a/', '/a/b/c') 
    ('b/c', '../..')
    
    >>> path_diff('a', 'a/b/c')
    ('b/c', '../..')

    >>> path_diff('a/x/y', 'a/b/c')
    ('../../b/c', '../../x/y')
    '''
    start = os.path.abspath(start)
    path = os.path.abspath(path)
    start_list = filter(None, start.split(os.path.sep)) # XXX(ot): is filter needed?
    path_list = filter(None, path.split(os.path.sep))
    
    prefix_list = os.path.commonprefix((start_list, path_list))
    r = len(prefix_list)
    s = len(start_list)
    p = len(path_list)

    s2p = os.path.sep.join([os.path.pardir]*(s-r) + path_list[r:])
    p2s = os.path.sep.join([os.path.pardir]*(p-r) + start_list[r:])

    return s2p, p2s

# XXX transitional, replace commands.getstatusoutput
def getstatusoutput(command):
    from subprocess import Popen, PIPE, STDOUT
    p = Popen(command, stdout=PIPE, stderr=STDOUT, shell=True)
    s = p.communicate()[0]
    return p.wait(), s



# -*- coding: utf-8 -*-
'''\
autobuild command
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

from optparse import make_option

from bpt import UserError
from bpt.ui.command import Command
from bpt.box import require_box
from bpt.autobuild import autobuild as autobuild_, UnsupportedPath

class autobuild(Command):
    '''Try to build a set of tarballs into the box, guessing the build commands'''

    usage_args = '<tarball> ...'

    def __init__(self):
        options = [make_option('-k', '--keep-intermediate', action='store_true',
                               dest='keep',
                               help='Do not delete intermediate files.'),
                   make_option('-c', '--configure_options', action='store',
                               dest='configure_options',
                               default='',
                               help='Options to pass to the configure/setup.py command line.')
                   ]
        Command.__init__(self, options)
        

    def _run(self, config, cmd_options, cmd_args):
        self._require_args(cmd_args, 1)
        require_box(config)

        try:
            for tarball in cmd_args:
                autobuild_(config.box, tarball, cmd_options.configure_options, cmd_options.keep)
        except UnsupportedPath, exc:
            raise UserError(str(exc))



# -*- coding: utf-8 -*-
'''\
Unit tests for box ui
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

import os
from tempfile import mktemp
from shutil import rmtree

from nose.tools import assert_raises, with_setup

from bpt.ui.main import main

def setup():
    global box1_path
    box1_path = mktemp()

def rmbox():
    try:
        rmtree(box1_path)
    except OSError:
        pass

def box(args):
    return main(['box'] + args, do_log=False)

@with_setup(teardown=rmbox)
def test_ui():
    assert box(['-b', box1_path, 'status']) != 0 # No box, error
    assert box(['create', box1_path]) == 0
    assert box(['create', box1_path]) != 0 # Box already existing

    assert box(['-b', box1_path, 'build', 'test/fake_sourcedirs/bar']) == 0
    assert box(['-b', box1_path, 'build', 'test/fake_sourcedirs/foo']) == 0
    assert box(['-b', box1_path, 'build', '-s', '-new', 'test/fake_sourcedirs/foo']) == 0

    assert box(['-b', box1_path, 'clean', '--deep', 'test/fake_sourcedirs/bar']) == 0
    assert box(['-b', box1_path, 'unittest', 'test/fake_sourcedirs/bar']) == 0
    
    assert box(['-b', box1_path, 'status']) == 0

    assert box(['-b', box1_path, 'disable', 'bar.*']) == 0
    assert box(['-b', box1_path, 'enable', 'bar.*']) == 0
    assert box(['-b', box1_path, 'disable', '-r', 'foo-0.1']) == 0

    assert box(['-b', box1_path, 'sync']) == 0

def test_help():
    assert box(['--help']) == 0
    assert box(['--help-commands']) == 0
    assert box(['unittest', '--help']) == 0
    
    
    



# -*- coding: utf-8 -*-
'''\
Basic setup.py for BPT using distutils
'''
#*****************************************************************************
#       Copyright (C) 2009 Giuseppe Ottaviano <giuott@gmail.com>
#
#  Distributed under the terms of the GPL License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************

from release_metadata import *

import os
from distutils.core import setup

f = open(os.path.join(os.path.dirname(__file__), 'README'))
long_description = f.read().strip()
f.close()

setup(name=app_name,
      version=app_version,
      description='Tool to create isolated environments',
      long_description=long_description,
      author='Giuseppe Ottaviano',
      author_email='giuott@gmail.com',
      url='http://pypi.python.org/pypi/bpt',
      packages=['bpt',
                'bpt.ui',
                'bpt.ui.commands'
                ],
      package_data={'bpt':['env_script.tmpl', 'bpt_base_script.sh']},
      scripts=['box', 'pip-box'],
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Console',
                   'Intended Audience :: Developers',
                   'Intended Audience :: System Administrators',
                   'License :: OSI Approved :: GNU General Public License (GPL)',
                   'Operating System :: POSIX',
                   'Topic :: Software Development :: Build Tools']
      )

